#ifndef __MORSE_GENERATOR
#define __MORSE_GENERATOR

#include "morse/model.js"
#include "morse/queue.js"
#include "util/tasker.js"

// Generate a stream of Morse in real-time.
Morse.Generator = function() {
  this.model = new Morse.Model();  // Morse timings and symbol table.

  this.message = new Morse.Queue();  // Stored Morse message.
  this.letter_queue = new Morse.Queue();  // Letters to be sent.
  this.symbol_queue = new Morse.Queue();  // Symbols to be sent in this letter.

  // One more class member, this.tasker, is defined at the end of this file.

  // Init() sets class members that might have callbacks and thus send data to
  // outlets, something that's not allowed during constuction.  You should call
  // Init() from a Max loadmsg.
  this.Init = function() {
    this.noteout = [64, 64];   // The MIDI note/velocity for Morse sound.
    this.midioffset = 64;      // The MIDI note that represents ASCII space.

    this.Clear();              // Clears the message variables.
  };

  // Clear() stops the generator and clears the message and queues.
  this.Clear = function() {
    this.Mode('stop');
    this.message.Clear();
    this.ClearQueues();
  };

  // Stops the generator and clears the queues.
  this.ClearQueues = function() {
    this.end_of_message = false;
    this.letter_queue.Clear();
    this.symbol_queue.Clear();
  };

  this.Mode = function(mode) {
    this.mode = mode;

    if (mode == 'stop')
      this.tasker.Stop();

    else if (mode == 'loop' && this.message.length)
      this.tasker.Run();

  };

  // NoteIn processes MIDI note on/offs from Live.
  this.NoteIn = function(pitch, velocity) {
    if (velocity) {  // A note-on.
      if (this.mode == 'gate')
        // The note becomes the template for our Morse notes.
        this.noteout = [pitch, velocity];

      if (this.mode == 'trigger') {
        this.ClearQueues();
        this.letter_queue.Set(this.message.queue);
        this.tasker.Run();
      }

    } else {  // A note-off.
      if (this.mode == 'gate') {
        this.noteout = [pitch, 0];
        this.makenote = ['stop'];
      }
    }
  };

  // KeyIn() converts keyboard entry to Morse.
  //
  // Input can either be a number, meaning an ASCII code, individual characters,
  // or strings.  Invalid characters will be displayed but silently ignored in
  // the Morse generation process.
  this.KeyIn = function(ascii) {
    var is_delete = (ascii == 0x8 || ascii == 0x7f);
    if (is_delete) {
      this.letter_queue.Delete();
      this.symbol_queue.Delete();
      this.message.Delete();

    } else {
      if (typeof(ascii) == 'number')
        ascii = String.fromCharCode(ascii);

      this.end_of_message = false;
      this.letter_queue.Append(ascii);
      if (this.mode == 'stream') {
        if (!this.tasker.running) {
          this.letter_queue.previous_character = '';
          this.tasker.Run();
        }

      } else {
        this.message.Append(ascii);
        if (this.mode == 'loop')
          this.tasker.Run();
      }
    }
  };

  // NextQueuedEvent() pops off the next item from the symbol_queue or
  // letter_queue and feeds it to the generator.
  this.NextQueuedEvent = function() {
    if (this.end_of_message) {
      this.end_of_message = false;
      this.symbol_queue.Clear();

      if (!this.letter_queue.length) {
        if (this.mode == 'loop'  && this.message.length) {
          this.letter_queue.Set(this.message.queue);
        } else {
          this.letter_queue.Clear();
          return;
        }
      }
    }

    // NextSymbol() and NextLetter() return true if processing needs to
    // continue during this cycle, or false if we're in the middle of a
    // note or a pause.
    while ((this.symbol_queue.length) ? this.NextSymbol() : this.NextLetter());
  };

  // Emits the next symbol from the symbol_queue.
  this.NextSymbol = function() {
    var symbol = this.symbol_queue.Pop();
    var on = (symbol == '.') ? this.model.dot : this.model.dash;
    var off = this.symbol_queue.length ? this.model.sep : this.model.gap;

    if (this.noteout[1])
      this.makenote = this.noteout.concat([this.model.AbsoluteTime(on)]);

    this.tasker.Schedule(this.model.AbsoluteTime(on + off));
    return false;
  };

  // Start the next letter from the letter_queue.
  this.NextLetter = function() {
    var letter = ' ';

    if (this.letter_queue.length > 0)
      letter = this.letter_queue.Pop();
    else
      this.end_of_message = true;

    var code = this.model.table.Code(letter);

    if (code == ' ') {
      this.symbol_queue.Clear();
      var time = this.model.AbsoluteTime(this.model.space - this.model.sep);
      this.tasker.Schedule(time);
      return false;

    } else if (code) {
      this.symbol_queue.Set(code);
    }

    return true;  // Start processing the next symbol, or failing that, letter.
  };

  // This actually runs the generator in a loop.
  this.tasker = new Tasker(this, this.NextQueuedEvent, []);
};

#endif  // __MORSE_GENERATOR
