#ifndef __MORSE_MODEL
#define __MORSE_MODEL

#include "morse/morse.js"
#include "morse/table.js"
#include "util/number_map.js"

// The class Morse.Model has settings for the production of Morse code.
//
// You can control the times of the various parts of the generator with the five
// components of Morse time - dot, dash, sep, gap and space - which are exposed
// as instance variables described below.
//
// Times are all relative to a base time unit.  The Morse class calculates the
// length of the standard reference Morse word "PARIS ", and then scales the
// base time unit to keep the correct number of words per minute.
//
// This means you can for example learn using the Farnsworth method (see
// http://en.wikipedia.org/wiki/Morse_code) by simply turning the value of "gap"
// and "sep" up a few notches - everything else will adjust properly.

Morse.Model = function() {
  this.table = new Morse.Table();  // Maps characters to Morse symbol groups.

  this.wpm = 20.0;  // Number of words per minute.

  // The five components of Morse time are public member values:
  this.dot = 1;     // Duration of a dot.
  this.dash = 3;    // Duration of a dash.
  this.sep = 1;     // Time between two symbols within a letter.
  this.gap = 3;     // Time between two letters within a word
  this.space = 7;   // Time between two words.

  // reference_word_map maps Morse component names to their frequency of occurence
  // in the standard reference word.
  this.reference_word_map = this.table.Measure('PARIS ');

  // AbsoluteTime() returns the real time value of one relative time unit, in
  // milliseconds.  For example, if you want to find out the actual amount of
  // time that 1 gap will take to execute
  this.AbsoluteTime = function(relative_time) {
    if (relative_time == undefined)
      relative_time = 1.0;

    // An instance of Morse.Model maps Morse component names to their relative
    // weights, and reference_word_map maps Morse component names to their
    // frequency in the standard reference word - so their dot product is the
    // total weight of the standard reference word with the current relative
    // time weights.
    var base_units = NumberMap.Dot(this.reference_word_map, this);
    return relative_time * (60000.0 / (this.wpm * base_units));
  };
};

#endif  // __MORSE_MODEL
